"""Contain the functions concerning the processing of data."""

import warnings
import numbers
import platform

import numpy as np
import scipy.sparse

from scipy.sparse import issparse
from scipy.sparse import dok_matrix
from scipy.sparse import lil_matrix

from sklearn.exceptions import DataConversionWarning
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_array, _assert_all_finite
from sklearn.preprocessing import normalize as skNormalize

from collections.abc import Sequence

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


def preprocess(X, centering=False, normalize=True, columns=True):
    """
        Preprocess features training data.

        Perform in-place centering or normalization, either of columns or rows
        of the input matrix X.

        Parameters
        ----------
            X (numpy array or scipy sparse CSR matrix):
                Input matrix

            centering (boolean) : default=False
                Perform a centering operation

            normalize (boolean): default=True
                l2-normalization

    """

    training_data_fortran = np.asfortranarray(X)

    if scipy.sparse.issparse(X):
        raise TypeError("The library does not supports sparse data.")

    if columns:
        if centering:
            column_means = np.mean(training_data_fortran, axis=0, keepdims=True)
            training_data_fortran = training_data_fortran - column_means
        if normalize:
            training_data_fortran = skNormalize(training_data_fortran, axis=0, norm="l2")
    else:
        if centering:
            row_means = np.mean(training_data_fortran, axis=1, keepdims=True)
            training_data_fortran = training_data_fortran - row_means
        if normalize:
            training_data_fortran = skNormalize(training_data_fortran, axis=1, norm="l2")

    return training_data_fortran


def sklearn_catch_warnings(y, check_y_kwargs):
    with warnings.catch_warnings():
        if not issparse(y):
            try:
                y = check_array(y, dtype=None, **check_y_kwargs)
            except (np.VisibleDeprecationWarning, ValueError) as e:
                if str(e).startswith("Complex data not supported"):
                    raise

                # dtype=object should be provided explicitly for ragged arrays,
                # see NEP 34
                y = check_array(y, dtype=object, **check_y_kwargs)


def sklearn_check_invalid_inputs(y):
    # Invalid inputs
    if y.ndim not in (1, 2):
        # Number of dimension greater than 2: [[[1, 2]]]
        return "unknown"
    if not min(y.shape):
        # Empty ndarray: []/[[]]
        if y.ndim == 1:
            # 1-D empty array: []
            return "binary"  # []
        # 2-D empty array: [[]]
        return "unknown"
    if not issparse(y) and y.dtype == object and not isinstance(y.flat[0], str):
        # [obj_1] and not ["label_1"]
        return "unknown"


def sklearn_check_old_format(y):
    # The old sequence of sequences format
    try:
        if (
            not hasattr(y[0], "__array__")
            and isinstance(y[0], Sequence)
            and not isinstance(y[0], str)
        ):
            raise ValueError(
                "Sequence of sequences are not"
                " supported; use a binary array or sparse"
                " matrix instead."
            )
    except IndexError:
        pass


# Code from scikit-learn
def type_of_target(y, input_name=""):
    """Determine the type of data indicated by the target.
    Note that this type is the most specific type that can be inferred.
    For example:
        * ``binary`` is more specific but compatible with ``multiclass``.
        * ``multiclass`` of integers is more specific but compatible with
          ``continuous``.
        * ``multilabel-indicator`` is more specific but compatible with
          ``multiclass-multioutput``.
    Parameters
    ----------
    y : {array-like, sparse matrix}
        Target values. If a sparse matrix, `y` is expected to be a
        CSR/CSC matrix.
    input_name: str, default=""
        The data name used to construct the error message.
        .. versionadded:: 1.1.0
    Returns
    -------
    target_type: str
        One of:
        * 'continuous': `y` is an array-like of floats that are not all
          integers, and is 1d or a column vector.
        * 'continuous-multioutput': `y` is a 2d array of floats that are
          not all integers, and both dimensions are of size > 1.
        * 'binary': `y` contains <= 2 discrete values and is 1d or a column
          vector.
        * 'multiclass': `y` contains more than two discrete values, is not a
          sequence of sequences, and is 1d or a column vector.
        * 'multiclass-multioutput': `y` is a 2d array that contains more
          than two discrete values, is not a sequence of sequences, and both
          dimensions are of size > 1.
        * 'multilabel-indicator': `y` is a label indicator matrix, an array
          of two dimensions with at least two columns, and at most 2 unique
          values.
        * 'unknown': `y` is array-like but none of the above, such as a 3d
          array, sequence of sequences, or an array of non-sequence objects.
    Examples
    --------
    >>> from sklearn.utils.multiclass import type_of_target
    >>> import numpy as np
    >>> type_of_target([0.1, 0.6])
    'continuous'
    >>> type_of_target([1, -1, -1, 1])
    'binary'
    >>> type_of_target(['a', 'b', 'a'])
    'binary'
    >>> type_of_target([1.0, 2.0])
    'binary'
    >>> type_of_target([1, 0, 2])
    'multiclass'
    >>> type_of_target([1.0, 0.0, 3.0])
    'multiclass'
    >>> type_of_target(['a', 'b', 'c'])
    'multiclass'
    >>> type_of_target(np.array([[1, 2], [3, 1]]))
    'multiclass-multioutput'
    >>> type_of_target([[1, 2]])
    'multilabel-indicator'
    >>> type_of_target(np.array([[1.5, 2.0], [3.0, 1.6]]))
    'continuous-multioutput'
    >>> type_of_target(np.array([[0, 1], [1, 1]]))
    'multilabel-indicator'
    """
    valid = (
        (isinstance(y, Sequence) or issparse(y) or hasattr(y, "__array__"))
        and not isinstance(y, str)
    )

    if not valid:
        raise ValueError(
            "Expected array-like (array or non-string sequence), got %r" % y
        )

    sparse_pandas = y.__class__.__name__ in ["SparseSeries", "SparseArray"]
    if sparse_pandas:
        raise ValueError("y cannot be class 'SparseSeries' or 'SparseArray'")

    if is_multilabel(y):
        return "multilabel-indicator"

    # DeprecationWarning will be replaced by ValueError, see NEP 34
    # https://numpy.org/neps/nep-0034-infer-dtype-is-object.html
    # We therefore catch both deprecation (NumPy < 1.24) warning and
    # value error (NumPy >= 1.24).
    check_y_kwargs = dict(
        accept_sparse=True,
        allow_nd=True,
        force_all_finite=False,
        ensure_2d=False,
        ensure_min_samples=0,
        ensure_min_features=0,
    )

    sklearn_catch_warnings(y, check_y_kwargs)

    sklearn_check_old_format(y)

    sklearn_check_invalid_inputs(y)

    # Check if multioutput
    if y.ndim == 2 and y.shape[1] > 1:
        suffix = "-multioutput"  # [[1, 2], [1, 2]]
    else:
        suffix = ""  # [1, 2, 3] or [[1], [2], [3]]

    # Check float and contains non-integer float values
    if y.dtype.kind == "f":
        # [.1, .2, 3] or [[.1, .2, 3]] or [[1., .2]] and not [1., 2., 3.]
        data = y.data if issparse(y) else y
        if np.any(data != np.floor(data)):
            _assert_all_finite(data)
            return "continuous" + suffix

    # Check multiclass
    first_row = y[0] if not issparse(y) else y.getrow(0).data
    if np.unique(y).shape[0] > 2 or (y.ndim == 2 and len(first_row) > 1):
        # [1, 2, 3] or [[1., 2., 3]] or [[1, 2]]
        return "multiclass" + suffix
    else:
        return "binary"  # [1, 2] or [["a"], ["b"]]


# Code from scikit-learn
def is_multilabel(y):
    """Check if ``y`` is in a multilabel format.
    Parameters
    ----------
    y : ndarray of shape (n_samples,)
        Target values.
    Returns
    -------
    out: bool
        Return ``True``, if ``y`` is in a multilabel format, else ```False``.
    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.utils.multiclass import is_multilabel
    >>> is_multilabel([0, 1, 0, 1])
    False
    >>> is_multilabel([[1], [0, 2], []])
    False
    >>> is_multilabel(np.array([[1, 0], [0, 0]]))
    True
    >>> is_multilabel(np.array([[1], [0], [0]]))
    False
    >>> is_multilabel(np.array([[1, 0, 0]]))
    True
    """
    if hasattr(y, "__array__") or isinstance(y, Sequence):
        # DeprecationWarning will be replaced by ValueError, see NEP 34
        # https://numpy.org/neps/nep-0034-infer-dtype-is-object.html
        check_y_kwargs = dict(
            accept_sparse=True,
            allow_nd=True,
            force_all_finite=False,
            ensure_2d=False,
            ensure_min_samples=0,
            ensure_min_features=0,
        )
        sklearn_catch_warnings(y, check_y_kwargs)

    if not (hasattr(y, "shape") and y.ndim == 2 and y.shape[1] > 1):
        return False

    if issparse(y):
        if isinstance(y, (dok_matrix, lil_matrix)):
            y = y.tocsr()
        labels = np.unique(y.data)
        return (
            len(y.data) == 0
            or (labels.size == 1 or (labels.size == 2) and (0 in labels))
            and (np.dtype.kind in "biu" or _is_integral_float(labels))  # bool, int, uint
        )
    else:
        labels = np.unique(y)

        return len(labels) < 3 and (
            y.dtype.kind in "biu" or _is_integral_float(labels)  # bool, int, uint
        )


# Code from scikit-learn
def _is_integral_float(y):
    return y.dtype.kind == "f" and np.all(y.astype(int) == y)


def check_labels(labels, estimator):
    """
    Verify the format of labels depending on the type of the estimator.

    Can convert labels in some cases.

    Parameters
    ----------
        labels (numpy array or scipy sparse CSR matrix):
            Numpy array containing labels

        estimator (ERM):
            The estimator which will be fitted

    Raises
    ------
        ValueError:
            Format of the labels does not respect the format supported by Cyanure classifiers.

        ValueError:
            Labels have an non finite value

        ValueError:
            Problem has only one class

    Returns
    -------
        labels (numpy array or scipy sparse CSR matrix):
            Converted labels if required by the estimator.

        label_encoder (sklearn.LabelEncoder):
            Convert text labels if needed
    """
    label_encoder = None

    if estimator.__sklearn_tags__().estimator_type == "classifier":

        if np.issubdtype(type(labels[0]), np.str_):
            label_encoder = LabelEncoder()
            label_encoder.fit(labels)
            labels = label_encoder.transform(labels)

        y_type = type_of_target(labels)
        if y_type not in [
            "binary",
            "multiclass"
        ]:
            raise ValueError("Unknown label type: %r" % y_type)

    else:
        if type(labels[0]) not in (np.float32, np.float64):
            logger.info("The labels have been converted to float64")
            labels = labels.astype('float64')

    _assert_all_finite(labels)

    if estimator.__sklearn_tags__().estimator_type == "classifier" and len(np.unique(labels)) == 1:
        raise ValueError("Classifier can't train when only one class is present.")

    return labels, label_encoder


def check_is_finite(array_to_test):
    if not np.any(np.isfinite(array_to_test)):
        raise ValueError(
            "Input contains NaN, infinity or a value too large for dtype('float64').")


def get_element(array):
    """
    Get an element from an array of any depth.

    Args
    ----
        array (Type of the element):
            Array we want to get an element

    Returns
    -------
        Type of the element:
            One of the element of the array
    """
    element = array[0]
    for i in range(len(array.shape) - 1):
        element = element[i]
    return element


def check_input_type(X, labels, estimator):
    """
    Verify the format of labels and features depending on the type of the estimator.

    Can convert labels in some cases.

    Parameters
    ----------
        X (numpy array or scipy sparse CSR matrix):
            Numpy array containing features

        labels (numpy array or scipy sparse CSR matrix):
            Numpy array containing labels

        estimator (ERM):
            The estimator which will be fitted

    Raises
    ------
        ValueError:
            Data are complex

        ValueError:
            Data contains non finite value

        TypeError:
            Sparsed features are not CSR

        TypeError:
            Sparsed labels are not CSR

    Returns
    -------
        X (numpy array or scipy sparse CSR matrix):
            Converted features if required by the estimator.

        labels (numpy array or scipy sparse CSR matrix):
            Converted labels if required by the estimator.

        label_encoder (sklearn.LabelEncoder):
            Convert text labels if needed
    """
    label_encoder = None

    if np.iscomplexobj(X) or np.iscomplexobj(labels):
        raise ValueError("Complex data not supported")

    if not scipy.sparse.issparse(X) and not scipy.sparse.issparse(labels):
        x_element = get_element(X)
        if type(x_element) not in (np.float32, np.float64):

            logger.info("The features have been converted in float64")
            X = X.astype('float64')

        labels, label_encoder = check_labels(labels, estimator)

        _assert_all_finite(X)

    else:
        if scipy.sparse.issparse(X):
            raise TypeError("The library does not supports sparse data.")

        X, labels = windows_conversion(X, labels)

    return X, labels, label_encoder


def windows_conversion(X, labels):

    if platform.system() == "Windows":
        if scipy.sparse.issparse(X):
            X.indptr = X.indptr.astype(np.float64).astype(np.intc)
            X.indices = X.indices.astype(np.float64).astype(np.intc)
        if scipy.sparse.issparse(labels):
            labels.indptr = labels.indptr.astype(np.float64).astype(np.intc)
            labels.indices = labels.indices.astype(np.float64).astype(np.intc)

    return X, labels


def check_positive_parameter(parameter, message):
    """
    Check that a parameter if a number and positive.

    Parameters
    ----------
        parameter (Any):
            Parameter to verify

        message (string):
            Message of the exception

    Raises
    ------
        ValueError:
            Parameter is not a number

        ValueError:
            Parameter is not positive
    """
    if not isinstance(parameter, numbers.Number):
        raise ValueError(message)

    if isinstance(parameter, numbers.Number) and parameter < 0:
        raise ValueError(message)


def check_parameters(estimator):
    """
    Verify that the different parameters of an estimator respect the constraints.

    Parameters
    ----------
        estimator (ERM):
            Estimator to veriffy
    """
    check_positive_parameter(
        estimator.tol, "Tolerance for stopping criteria must be positive")
    check_positive_parameter(estimator.max_iter,
                             "Maximum number of iteration must be positive")
    check_positive_parameter(estimator.lambda_1,
                             "Penalty term must be positive")

    instance_class = estimator.__class__
    instance = instance_class()

    # Verify that it is not the default value
    if (estimator.penalty is None or estimator.penalty == "none") and estimator.lambda_1 != instance.lambda_1:
        warnings.warn("Setting penalty='none' will ignore the lambda_1")


def convert_to_array(X, labels):
    if not scipy.sparse.issparse(X) and not scipy.sparse.issparse(labels):
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(labels, np.ndarray):
            labels = np.array(labels)

    return X, labels


def check_input_fit(X, labels, estimator):
    """
    Check the different input arrays required for training according to the estimator type.

    Can convert data if necessary.

    Parameters
    ----------
        X (numpy array or scipy sparse CSR matrix):
            Numpy array containing features

        labels (numpy array or scipy sparse CSR matrix):
            Numpy array containing labels

        estimator (ERM):
            The estimator which will be fitted

    Raises
    ------
        ValueError:
            There is only one feature.

        ValueError:
            There is no sample.

        ValueError:
            An observation has no label.

        ValueError:
            Feature array has no feature

        ValueError:
            Features and labels does not have the same number of observations.

        ValueError:
            There is only one sample.

    Returns
    -------
        X (numpy array or scipy sparse CSR matrix):
            Converted features if required by the estimator.

        labels (numpy array or scipy sparse CSR matrix):
            Converted labels if required by the estimator.

        label_encoder (sklearn.LabelEncoder):
            Convert text labels if needed
    """

    X, labels = convert_to_array(X, labels)

    if X.ndim == 1:
        raise ValueError("The training array has only one dimension.")

    if X.shape[0] == 0:
        raise ValueError("Empty training array")

    if labels is None or True in np.array(np.equal(labels, None)):
        raise ValueError("y should be a 1d array")

    if len(X.shape) > 1 and X.shape[1] == 0:
        raise ValueError("0 feature(s) (shape=(" + str(X.shape[0]) + ", 0)) while a minimum of "
              + str(X.shape[0]) + " is required.")

    if labels.shape[0] != X.shape[0]:
        raise ValueError(
            "X and labels should have the same number of observations")

    if X.shape[0] == 1:
        raise ValueError("There should have more than 1 sample")

    if not estimator._get_tags()["multioutput"] and \
       not estimator._get_tags()["multioutput_only"] and labels.ndim > 1:
        warnings.warn(
            "A column-vector y was passed when a 1d array was expected", DataConversionWarning)

    X, labels, label_encoder = check_input_type(X, labels, estimator)
    check_parameters(estimator)

    return X, labels, label_encoder


def check_input_inference(X, estimator):
    """
    Check the format of the array which will be used for inference. Input array can be converted.

    Parameters
    ----------
        X (numpy array or scipy sparse CSR matrix):
            Array which will be used for inference

        estimator (ERM):
            Estimator which will be used

    Raises
    ------
        ValueError:
            One of the value is not finite

        ValueError:
            Shape of features is not correct

        ValueError:
            Shape of features does not correspond to estimators shape

    Returns
    -------
        X (numpy array or scipy sparse CSR matrix):
            Potentially converted array (if converted as numpy.float64)

    """
    if not scipy.sparse.issparse(X):
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        x_element = get_element(X)
        if type(x_element) not in (np.float32, np.float64):
            X = X.astype("float64")

        _assert_all_finite(X)

    if X.ndim == 1:
        raise ValueError("Reshape your data")

    if X.shape[1] != estimator.n_features_in_:
        raise ValueError(f"X has 1 features, but estimator is expecting {estimator.n_features_in_} features as input")

    return X



from cyanure_gpu.logger import setup_custom_logger
import torch
from typing import Tuple
import abc

from cyanure_gpu.losses.loss import Loss
from cyanure_gpu.constants import DEVICE

logger = setup_custom_logger("INFO")


class LinearLossMat(Loss):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)
        self.ones = torch.ones(self.input_data.size(dim=1)).to(DEVICE)

    def add_grad(self, input: torch.Tensor, i: int, a: float = 1.0) -> torch.Tensor:
        sgrad = self.scal_grad(input, i)
        return self.add_dual_pred(i, sgrad, a)

    def double_add_grad(self, input1: torch.Tensor, input2: torch.Tensor, i: int,
                        eta1: float = 1.0, eta2: float = -1.0, dummy: float = 1.0) -> torch.Tensor:
        sgrad1 = self.scal_grad(input1, i)
        sgrad2 = self.scal_grad(input2, i)
        sgrad1 = sgrad2 * eta2 + sgrad1 * eta1
        return self.add_dual_pred(i, sgrad1)

    def transpose(self) -> bool:
        return True

    def add_feature_tensor(self, input: torch.Tensor, input2: torch.Tensor, s: float) -> torch.Tensor:
        return self.add_dual_pred_tensor(input, input2, s, 1.0)

    def add_feature(self, i: int, s: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        return self.add_dual_pred(i, s, input2, 1.0, 1.0)

    # _X  is  p x n
    # input is nclass x p
    # output is nclass x n
    def pred_tensor(self, input: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        if (self.intercept):
            weight, bias = self.get_wb(input)
            if input2 is not None:
                output = torch.matmul(weight, self.input_data) + input2
            else:
                output = torch.matmul(weight, self.input_data)
            output = output + torch.outer(bias, self.ones)
        else:
            if input2 is not None:
                output = torch.matmul(input, self.input_data) + input2
            else:
                output = torch.matmul(input, self.input_data)
        return output

    def pred(self, ind: int, input: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        col = self.input_data[:, ind]
        if (self.intercept):
            weight, bias = self.get_wb(input)
            output = torch.matmul(weight, col) + input2
            output = output + self.scale_intercept * bias
        else:
            if input2 is not None:
                output = torch.matmul(input, col) + input2
            else:
                output = torch.matmul(input, col)

        return output

    def add_dual_pred_tensor(self, input: torch.Tensor, input2: torch.Tensor, a1: float = 1.0, a2: float = 1.0) -> torch.Tensor:
        if (self.intercept):
            output = torch.zeros(input.size(dim=0), self.input_data.size(dim=0) + 1).to(DEVICE)
            weight, bias = self.get_wb(input2)
            #  W = input * X.T =  (X* input.T).T
            weight = a2 * weight + a1 * torch.matmul(input, torch.transpose(self.input_data, 0, 1))
            bias = bias * a2 + a1 * torch.matmul(input, self.ones)
            bias = torch.unsqueeze(bias, 1)
            output = torch.cat((weight, bias), dim=1)
        else:
            if input2 is not None:
                output = a2 * input2 + a1 * torch.matmul(input, self.input_data.t())
            else:
                output = a1 * torch.matmul(input, self.input_data.t())

        return output

    def add_dual_pred(self, ind: int, input: torch.Tensor, input2: torch.Tensor, a: float = 1.0, b: float = 1.0) -> torch.Tensor:
        col = self.input_data[:, ind]
        if (b != 1.0):
            input2 *= b
        if (self.intercept):
            output = torch.zeros(input.size(dim=1), self.input_data.size(dim=0) + 1)
            # Weight and bias are zeros
            weight, bias = self.get_wb(input)
            weight = weight + a * torch.outer(input, col)
            bias = bias + a*self.scale_intercept * input
            output = torch.cat((weight, bias), dim=1)
        else:
            if input2 is not None:
                output = input2 + a * self.scale_intercept * input
            else:
                output = a * self.scale_intercept * input

        return output

    def get_wb(self, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        p = input.size(dim=1)
        weight = input[:, : p-1]
        bias = input[:, p-1]

        return weight, bias

    @abc.abstractmethod
    def scal_grad(self, input: torch.Tensor, i: int) -> torch.Tensor:
        return


class LossMat(LinearLossMat):

    def __init__(self, loss: Loss):
        self.loss_list = list()
        self.data_list = list()
        self.n = loss.labels.size(dim=1)
        self.num_class = loss.labels.size(dim=0)
        self.labels_transpose = torch.transpose(loss.labels, 0, 1)
        for i in range(self.num_class):
            self.data_list[i] = loss.input_data
            ycol = self.labels_transpose[:, i]
            self.loss_list[i] = type(loss)(self.data_list[i], ycol, loss.intercept)
        self.id = self.loss_list[0].id

    def eval_tensor(self, input_tensor: torch.Tensor) -> float:
        sum_value = torch.tensor(0.0)  # Initialize as a PyTorch tensor

        for ii in range(self.num_class):  # Assuming _N is defined somewhere
            col = input_tensor[:, ii]
            value = self.loss_list[ii].eval(col)

            sum_value += value

        return sum_value

    def eval(self, input_tensor: torch.Tensor, i: int) -> float:
        sum_value = torch.tensor(0.0)  # Initialize as a PyTorch tensor

        for ii in range(self.num_class):
            col = input_tensor[:, ii]
            value = self.loss_list[ii].eval(col, i)

            sum_value += value

        return sum_value

    def add_grad(self, input_tensor: torch.Tensor, i: int, eta: float = 1.0) -> torch.Tensor:
        output = torch.zeros_like(input_tensor)

        for ii in range(self.num_class):
            output[:, ii] = self.loss_list[ii].add_grad(input_tensor[:, ii], i, eta)

        return output

    def double_add_grad(self, input1: torch.Tensor, input2: torch.Tensor, i: int,
                        eta1: float = 1.0, eta2: float = -1.0, dummy: float = 1.0) -> torch.Tensor:
        # Input tensor is a vector
        output = torch.zeros(input1.size())

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            output[:, ii] = self.loss_list[ii].double_add_grad(input1[:, ii], input2[:, ii], i, eta1, eta2, dummy)

        return output

    def grad(self, input_tensor: torch.Tensor) -> torch.Tensor:
        output = torch.zeros_like(input_tensor)

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            output[:, ii] = self.loss_list[ii].grad(input_tensor[:, ii])

        return output

    def print(self) -> None:
        logger.info("Loss for matrices")
        self.loss_list[0].print()

    def provides_fenchel(self) -> bool:
        return self.loss_list[0].provides_fenchel()

    def get_dual_variable(self, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        # Input tensor is a vector
        grad1 = torch.zeros((self.n, input.size(dim=1)))
        grad2 = torch.zeros(input.size())

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            grad1[:, ii], grad2[:, ii] = self.loss_list[ii].get_dual_variable(input[:, ii])

        return grad1, grad2

    def fenchel(self, input_tensor: torch.Tensor) -> float:
        sum_value = torch.tensor(0.0)  # Initialize as a PyTorch tensor

        for ii in range(self.num_class):
            col = input_tensor[:, ii]
            value = self.loss_list[ii].fenchel(col)

            sum_value.add_(value)

        return sum_value

    def lipschitz(self) -> float:
        return self.loss_list[0].lipschitz()

    def lipschitz_li(self, Li: torch.Tensor) -> torch.Tensor:
        return self.loss_list[0].lipschitz_li(Li)

    # input; nclass x n
    # output: p x nclass
    def add_feature_tensor(self, input: torch.Tensor, input2: torch.Tensor, s: float) -> torch.Tensor:
        output = torch.zeros(input.size())

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            output[:, ii] = self.loss_list[ii].add_feature_tensor(input[ii, :], input2[:, ii], s)

        return output

    def add_feature(self, i: int, s: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        # Input tensor is a vector
        output = torch.zeros((self.loss_list[0].input_data.size(dim=0), self.num_class))

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            output[:, ii] = self.loss_list[ii].add_feature(i, s[ii], input2[:, ii])

        return output

    def scal_grad(self, input_tensor: torch.Tensor, i: int) -> torch.Tensor:
        output = torch.zeros(self.num_class)

        for ii in range(self.num_class):  # Assuming self.num_class is defined somewhere
            output[ii] = self.loss_list[ii].scal_grad(input_tensor[:, ii], i)

        return output

    def transpose(self) -> bool:
        return False

    def get_grad_aux(self, input: torch.Tensor) -> None:
        logger.error("Not used")

    def lipschitz_constant(self) -> float:
        logger.error("Not used")
        return 0

    def get_dual_constraints(self, grad1: torch.Tensor) -> None:
        logger.error("Not used")



import torch
import torch.nn

from cyanure_gpu.erm.erm import Estimator
from cyanure_gpu.erm.param.model_param import ModelParameters
from cyanure_gpu.erm.param.problem_param import ProblemParameters
from cyanure_gpu.logger import setup_custom_logger
from cyanure_gpu.losses.logistic import LogisticLoss
from cyanure_gpu.losses.square import SquareLoss
from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.regularizers.ridge import Ridge
from cyanure_gpu.regularizers.lasso import Lasso
from cyanure_gpu.regularizers.none import NoRegul
from cyanure_gpu.solvers.ista import ISTA_Solver
from cyanure_gpu.constants import EPSILON

from typing import Tuple

logger = setup_custom_logger("INFO")


class SimpleErm(Estimator):

    def __init__(self, initial_weight: torch.Tensor, weight: torch.Tensor, problem_parameters: ProblemParameters,
                 model_parameters: ModelParameters, optim_info: torch.Tensor, dual_variable: torch.Tensor):
        super().__init__(problem_parameters, model_parameters, optim_info)
        self.initial_weight = initial_weight
        self.weight = weight
        self.dual_variable = dual_variable

    def solve_problem(self, features: torch.Tensor, labels: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:

        if (self.model_parameters.verbose):
            logger.info("Matrix X, n=" + str(features.size(dim=1)) + ", p=" + str(features.size(dim=0)))

        self.verify_input(features)
        loss = self.get_loss(features, labels)
        regul = self.get_regularization()
        solver = None

        if (self.model_parameters.max_iter == 0):
            parameter_tmp = self.model_parameters
            parameter_tmp.verbose = False
            solver = ISTA_Solver(loss,  regul, parameter_tmp, False)
            solver.eval(self.initial_weight)
            self.weight = torch.clone(self.initial_weight)
        else:

            if (solver is None):
                regul.strong_convexity()
                solver = self.get_solver(loss, regul, self.model_parameters)

            if (solver is None):
                self.weight = torch.clone(self.initial_weight)
                return None

            if (self.problem_parameters.intercept):
                new_initial_weight = loss.set_intercept(self.initial_weight)
            else:
                new_initial_weight = torch.clone(self.initial_weight)

            if (self.dual_variable is not None and self.dual_variable.size(dim=0) != 0):
                solver.set_dual_variable(self.dual_variable)

            self.weight, fprox = solver.solve(new_initial_weight, self.weight)

            if (self.problem_parameters.intercept):
                self.weight = loss.reverse_intercept(self.weight)

        if (self.problem_parameters.regul == "L1"):
            self.weight[torch.abs(self.weight) < EPSILON] = 0

        if (self.weight is None):
            self.weight = self.initial_weight

        return solver.get_optim_info(), self.weight

    def verify_input(self, X: torch.Tensor) -> None:
        if (self.problem_parameters.intercept):
            if (X.size(dim=0) + 1 != self.initial_weight.size(dim=0)):
                logger.error("Dimension of initial point is not consistent."
                             " With intercept, if X is m x n, w0 should be (n+1)-dimensional.")
                return None
        else:
            if (X.size(dim=0) != self.initial_weight.size(dim=0)):
                logger.error("Dimension of initial point is not consistent. If X is m x n, w0 should be n-dimensional.")
                return None

        if (self.model_parameters.max_iter < 0):
            raise ValueError("Maximum number of iteration must be positive")
        if (self.problem_parameters.lambda_1 < 0):
            raise ValueError("Penalty term must be positive")
        if (self.model_parameters.tol < 0):
            raise ValueError("Tolerance for stopping criteria must be positive")

    def get_regularization(self) -> Regularizer:
        regul = None

        self.problem_parameters.regul = self.problem_parameters.regul.upper()
        if self.problem_parameters.regul == "L2":
            regul = Ridge(self.problem_parameters)
        elif self.problem_parameters.regul == "L1":
            regul = Lasso(self.problem_parameters)
        else:
            logger.error("Not implemented, no regularization is chosen")
            regul = NoRegul(self.problem_parameters)

        return regul

    def get_loss(self, data: torch.Tensor, y: torch.Tensor):
        loss = None

        self.problem_parameters.loss = self.problem_parameters.loss.upper()
        if (self.problem_parameters.loss == "SQUARE"):
            loss = SquareLoss(data, y, self.problem_parameters.intercept)
        elif (self.problem_parameters.loss == "LOGISTIC"):
            loss = LogisticLoss(data, y, self.problem_parameters.intercept)
        else:
            logger.error("Not implemented, square loss is chosen by default")
            loss = SquareLoss(data, y, self.problem_parameters.intercept)

        return loss



import torch

from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class Ridge(Regularizer):

    def __init__(self, model: ProblemParameters):
        super().__init__(model)

        self.id = "L2"

    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        output = torch.clone(input)

        output = output / (1.0 + self.lambda_1 * eta)
        if (self.intercept):
            if len(output.shape) == 1:
                n = input.size(dim=0)
                output[n - 1] = input[n - 1]
            else:
                p = input.size(dim=1)
                output[:, p - 1] = input[:, p-1]
        return output

    def eval_tensor(self, input: torch.Tensor) -> float:
        n = input.size(dim=0)
        res = torch.sum(torch.tensordot(input, input, dims=len(input.shape)))
        return (0.5 * self.lambda_1 * (res - torch.pow(input[n - 1], 2)) if self.intercept else 0.5 * self.lambda_1 * res)

    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> float:
        if (self.intercept and (abs(grad2[grad2.size(dim=0) - 1]) > 1e-6)):
            value = float("inf")
        else:
            value = self.eval_tensor(grad2) / (torch.pow(self.lambda_1, 2))
        return value, grad1, grad2

    def print(self) -> None:
        logger.info(self.getName())

    def strong_convexity(self) -> float:
        return 0 if self.intercept else self.lambda_1

    def lazy_prox(self, input: torch.Tensor, indices: torch.Tensor, eta: float) -> None:
        scal = 1.0 / (1.0 + self.lambda_1 * eta)
        output = torch.zeros(input.size())
        p = input.size(dim=0)
        r = indices.size(dim=0)
        for jj in range(r):
            output[indices[jj]] = scal * input[indices[jj]]
        if (self.intercept):
            if len(output.shape) == 1:
                n = input.size(dim=0)
                output[n - 1] = input[n - 1]
            else:
                p = input.size(dim=1)
                output[:, p - 1] = input[:, p-1]

        return output

    def is_lazy(self) -> bool:
        return True

    def getName(self) -> str:
        return "L2 regularization"



import torch

from cyanure_gpu.losses.loss_matrix import LinearLossMat
from cyanure_gpu.constants import DEVICE
from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class MultiClassLogisticLoss(LinearLossMat):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)
        self.n_classes = int(torch.max(y) + 1)
        self.id = 'MULTI_LOGISTIC'
        self.number_data = self.labels.shape[0]
        self.one_hot = torch.nn.functional.one_hot(self.labels.to(torch.int64), self.n_classes).T.to(torch.int32)
        self.boolean_mask = torch.zeros(self.n_classes, self.number_data, dtype=torch.bool).to(DEVICE)
        index_mask = torch.arange(self.n_classes).unsqueeze(1).expand(self.n_classes, self.number_data).to(DEVICE)
        label_mask = torch.unsqueeze(self.labels, 0).expand(self.n_classes, self.number_data)
        self.boolean_mask = torch.eq(index_mask, label_mask)
        self.loss_labels = self.labels.type(torch.LongTensor).to(DEVICE)

    def pre_compute(self, input: torch.Tensor) -> torch.Tensor:
        # Get prediction tensor
        tmp = self.pred_tensor(input, None)

        # Compute the difference with masked select and broadcasting
        diff = torch.masked_select(tmp, self.boolean_mask).unsqueeze(0).expand(self.n_classes, self.number_data)

        # Subtract difference (avoid in-place operations)
        tmp = tmp - diff

        # Apply log-sum-exp trick to improve numerical stability
        mm = tmp.max(dim=0, keepdim=True).values
        tmp = tmp - mm
        tmp = tmp.exp()

        # Sum matrix along the first dimension, ensuring numerical stability with epsilon
        sum_matrix = tmp.sum(dim=0, keepdim=True)

        return tmp, sum_matrix, mm

    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> float:
        if precompute is None:
            if matmul_result is not None:
                tmp = matmul_result
            else:
                tmp = self.pred_tensor(input, None)

            diff = torch.masked_select(tmp, self.boolean_mask).unsqueeze(0).expand(self.n_classes, self.number_data)

            tmp = tmp - diff
            # Find max and perform subsequent operations

            mm_vector = tmp.max(dim=0, keepdim=True).values
            tmp.sub_(mm_vector)

            tmp = tmp.exp()

            sum_matrix = tmp.sum(dim=0, keepdim=True)
        else:
            tmp = precompute[0]
            sum_matrix = precompute[1]
            mm_vector = precompute[2]

        log_sums = torch.log(sum_matrix)
        sum_value = torch.sum(mm_vector) + torch.sum(log_sums)

        return sum_value / self.number_data

    def eval(self, input: torch.Tensor, i: int) -> float:
        tmp = self.pred(i, input)
        tmp += -tmp[self.labels[i]]
        mm = torch.max(tmp)
        tmp += -mm
        tmp = torch.exp(tmp)
        return mm + torch.log(torch.sum(tmp))

    def print(self) -> None:
        logger.info("Multiclass logistic Loss is used")

    def xlogx(self, x):

        # Handling condition where x is greater than or equal to 1e-20
        res = x * torch.log(x)

        neg_mask = x < -1e-20
        # Handling condition where x is less than -1e-20
        res.masked_fill_(neg_mask, float('inf'))

        small_mask = x < 1e-20
        # Handling condition where x is less than 1e-20
        res.masked_fill_(small_mask, 0)

        return res

    def fenchel(self, input: torch.Tensor) -> float:
        n = input.size(1)

        # Use advanced indexing to select the relevant elements
        # Add the condition for sum += xlogx(input[i * _nclasses + j] + 1.0)
        input += self.one_hot

        selected_xlogx = self.xlogx(input)
        # Sum along the second dimension (class dimension)
        sum_val = torch.sum(selected_xlogx)

        return sum_val / n

    def get_grad_aux2(self, col: torch.Tensor, ind: int) -> torch.Tensor:
        value = col[ind].clone()
        col = col - value
        mm = torch.max(col)
        col = col - mm
        col = torch.exp(col)
        col = col * (1 / (torch.sum(torch.abs(col))))

        col[ind] = 0
        col[ind] = -(torch.sum(torch.abs(col)))
        return col

    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None,
                     precompute: torch.Tensor = None) -> torch.Tensor:

        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)

            diff = torch.masked_select(grad1, self.boolean_mask).unsqueeze(0).expand(self.n_classes, self.number_data)
            grad1 = grad1 - diff

            # Apply log-sum-exp trick
            mm = grad1.max(dim=0, keepdim=True).values
            grad1 = (grad1 - mm).exp()

            sum_matrix = grad1.sum(dim=0, keepdim=True)
        else:
            grad1 = precompute[0]
            sum_matrix = precompute[1]

        grad1 = grad1 / (sum_matrix)  # More stable division

        # Apply the mask to zero out certain elements
        grad1 = torch.where(self.one_hot.bool(), torch.tensor(0.0, device=grad1.device, dtype=grad1.dtype), grad1)

        # Compute the sum of absolute values along the first dimension
        abs_sum = torch.sum(torch.abs(grad1), dim=0, keepdim=True)

        # Compute the adjustment tensor
        adjustment = self.one_hot.float() * abs_sum

        # Subtract the adjustment tensor from grad1
        grad1.sub_(adjustment)

        return grad1

    def get_grad_aux_to_compile(self, matmul_result: torch.Tensor) -> torch.Tensor:
        grad1 = matmul_result

        # Subtract one-hot encoded vector from each element in grad1
        diff = grad1[self.labels.to(torch.int64), torch.arange(grad1.shape[1])]
        grad1 = grad1.clone()
        grad1 -= diff

        grad1.sub_(diff)
        # Find max and perform subsequent operations

        mm = grad1.max(dim=0, keepdim=True).values
        grad1.sub_(mm)

        grad1 = grad1.exp()

        sum_matrix = torch.abs(grad1).sum(dim=0, keepdim=True)

        grad1 /= sum_matrix

        # Compute the mask for elements to be zeroed out
        mask = 1 - self.one_hot

        # Apply the mask to grad1
        grad1 *= mask

        # Compute the sum of absolute values along the first dimension
        abs_sum = torch.sum(torch.abs(grad1), dim=0, keepdim=True)

        # Compute the adjustment tensor
        adjustment = self.one_hot * abs_sum

        # Subtract the adjustment tensor from grad1
        grad1.sub_(adjustment)

        return grad1

    def scal_grad(self, input: torch.Tensor, i: int) -> torch.Tensor:
        col = self.pred(i, input, None)
        return self.get_grad_aux2(col, int(self.labels[i]))

    def lipschitz_constant(self) -> float:
        return 0.25

    def get_dual_constraints(self, grad1: torch.Tensor) -> torch.Tensor:
        if self.intercept:
            for i in range(grad1.size(0)):
                grad1[i, :] = self.project_sft(grad1[i, :], self.labels, i)
        return grad1

    def project_sft(self, grad1_vector: torch.Tensor, labels: torch.Tensor, clas: int) -> torch.Tensor:
        labels_binary = torch.where(labels == clas, 1.0, -1.0)
        return self.project_sft_binary(grad1_vector, labels_binary)

    def project_sft_binary(self, grad1: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        mean = torch.mean(grad1)

        if mean > 0:
            ztilde = grad1 + torch.where(y > 0, 1.0, 0.0)
            count = torch.sum(y > 0).item()
            xtilde = self.l1project(ztilde, count)
            grad1 = xtilde - torch.where(y > 0, 1.0, 0.0)
        else:
            ztilde = torch.where(y > 0, -grad1, -grad1 + 1.0)
            count = torch.sum(y <= 0).item()
            xtilde = self.l1project(ztilde, count)
            grad1 = torch.where(y > 0, -xtilde, -xtilde + 1.0)

        return grad1

    def l1project(self, input: torch.Tensor, thrs: float, simplex: bool = False) -> torch.Tensor:
        if simplex:
            output = torch.clamp(input, min=0)
        else:
            output = input.abs()

        norm1 = torch.sum(output)
        if norm1 <= thrs:
            return input if not simplex else output

        # Sort the input tensor in descending order
        sorted_output, _ = torch.sort(output, descending=True)

        # Calculate the cumulative sum
        cumulative_sum = torch.cumsum(sorted_output, dim=0) - thrs

        # Find rho, which is the largest index where the condition holds
        tmp = sorted_output * torch.arange(1, sorted_output.size(0) + 1, device=input.device) > cumulative_sum
        rho = torch.nonzero(tmp, as_tuple=True)[0].max()

        # Calculate the threshold lambda
        lambda_1 = cumulative_sum[rho] / (rho + 1)

        # Threshold the input tensor
        output = input.sign() * torch.clamp(output - lambda_1, min=0)

        return output



import os

# Override sdist to always produce .zip archive
from distutils.command.sdist import sdist as _sdist

from setuptools import setup, find_packages

class sdistzip(_sdist):
    def initialize_options(self):
        _sdist.initialize_options(self)
        self.formats = ['zip', 'gztar']

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'VERSION')) as version_file:
    version = version_file.read().strip()

setup(name='cyanure_gpu',
      version=version,
      author="Julien Mairal",
      author_email="julien.mairal@inria.fr",
      license='bsd-3-clause',
      url="https://inria-thoth.github.io/cyanure_gpu/welcome.html",
      description='optimization toolbox for machine learning',
      install_requires=['scikit-learn', 'torch<=2.5.1', 'numpy==1.26.4'],
      packages=find_packages(),
      cmdclass={'sdist': sdistzip},
      long_description="Cyanure is an open-source Python software package. This is a partial re implementation of Cyanure package. It provides a simple Python API, which should be fully compatible with scikit-learn.")




import torch

from cyanure_gpu.losses.loss import LinearLossVec

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class LogisticLoss(LinearLossVec):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)
        self.id = "LOGISTIC"

    def eval(self, input: torch.Tensor, i: int) -> float:
        res = self.labels[i] * self.pred(i, input)
        if res > 0:
            return torch.log(1.0 + torch.exp(-res))
        else:
            return torch.log(1.0 + torch.exp(res))

    def pre_compute(self, input: torch.Tensor) -> float:

        tmp = self.pred_tensor(input, None)

        tmp = torch.mul(tmp, self.labels)

        return tmp

    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> float:
        if precompute is None:
            if matmul_result is None:
                tmp = self.pred_tensor(input, None)
            else:
                tmp = matmul_result.clone()
            tmp = torch.mul(tmp, self.labels)
        else:
            tmp = precompute
        tmp = torch.log(1.0 + torch.exp(torch.neg(tmp)))
        return torch.sum(tmp) / (tmp.size(dim=0))

    def print(self) -> None:
        logger.info("Logistic Loss is used")

    def fenchel(self, input: torch.Tensor) -> float:

        n = input.size(dim=0)
        prod = torch.mul(self.labels, input)
        sum_vector = torch.special.xlogy(1.0+prod, 1.0+prod)+torch.special.xlogy(-prod, -prod)
        return torch.sum(sum_vector)/n

    def scal_grad(self, input: torch.Tensor, i: int) -> float:
        label = self.labels[i]
        ss = self.pred(i, input)
        s = -label/(1.0+torch.exp(label*ss))

        return s

    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None,
                     precompute: torch.Tensor = None) -> torch.Tensor:
        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)
            grad1 = torch.mul(grad1, self.labels)
        else:
            grad1 = precompute

        grad1 = 1.0 / (torch.exp(grad1) + 1.0)
        grad1 = torch.mul(grad1, self.labels)
        grad1 = torch.neg(grad1)

        return grad1

    def lipschitz_constant(self) -> float:
        return 0.25

    def get_dual_constraints(self, grad1: torch.Tensor) -> torch.Tensor:
        if (self.intercept):
            grad1 = self.project_sft_binary(grad1, self.labels)

        return grad1

    def project_sft_binary(self, grad1: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        mean = torch.mean(grad1)
        if mean > 0:
            ztilde = grad1 + torch.where(y > 0, 1.0, 0.0)
            count = torch.sum(y > 0).item()
            xtilde = self.l1project(ztilde, count)
            grad1 = xtilde - torch.where(y > 0, 1.0, 0.0)
        else:
            ztilde = torch.where(y > 0, -grad1, -grad1 + 1.0)
            count = torch.sum(y <= 0).item()
            xtilde = self.l1project(ztilde, count)
            grad1 = torch.where(y > 0, -xtilde, -xtilde + 1.0)

        return grad1

    def l1project(self, input: torch.Tensor, thrs: float, simplex: bool = False) -> torch.Tensor:

        if simplex:
            output = torch.clamp(input, min=0)
        else:
            output = input.abs()

        norm1 = torch.sum(output)
        if norm1 <= thrs:
            return input if not simplex else output

        # Sort the input tensor in descending order
        sorted_output, _ = torch.sort(output, descending=True)

        # Calculate the cumulative sum
        cumulative_sum = torch.cumsum(sorted_output, dim=0) - thrs

        # Find rho, which is the largest index where the condition holds
        tmp = sorted_output * torch.arange(1, sorted_output.size(0) + 1, device=input.device) > cumulative_sum
        rho = torch.nonzero(tmp, as_tuple=True)[0].max()

        # Calculate the threshold lambda
        lambda_1 = cumulative_sum[rho] / (rho + 1)

        # Threshold the input tensor
        output = input.sign() * torch.clamp(output - lambda_1, min=0)

        return output



"""Contain the logger configuration used in the project."""

import logging

LEVEL_CORRESPONDANCE = {'CRITICAL': 50, 'ERROR': 40, 'WARNING': 30, 'INFO': 20, 'DEBUG': 10}

LOGGER = None


def setup_custom_logger(level):
    """
    Init the logger for the application.

    :param level: Level of the expected log
    :return: Logger instance
    """
    global LOGGER

    if LOGGER is None:
        formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        LOGGER = logging.getLogger("Cyanure_gpu")
        LOGGER.setLevel(LEVEL_CORRESPONDANCE[level])
        LOGGER.addHandler(handler)
    return LOGGER



import torch

from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class Lasso(Regularizer):

    def __init__(self, model: ProblemParameters):
        super().__init__(model)

        self.id = "L1"

    def fastSoftThrs(self, x: float) -> float:
        return x + 0.5*(torch.abs(x - self.lambda_1) - torch.abs(x + self.lambda_1))

    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        output = input + 0.5 * (torch.abs(input - eta * self.lambda_1) - torch.abs(input + eta * self.lambda_1))

        if (self.intercept):
            if len(output.shape) == 1:
                n = input.size(dim=0)
                output[n - 1] = input[n - 1]
            else:
                p = input.size(dim=1)
                output[:, p - 1] = input[:, p-1]
        return output

    def eval_tensor(self, input: torch.Tensor) -> float:
        n = input.size(dim=0)
        res = torch.sum(torch.abs(input))
        return (self.lambda_1 * (res - torch.abs(input[n - 1])) if self.intercept else self.lambda_1 * res)

    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> float:
        indices = (torch.abs(grad2) == torch.max(torch.abs(grad2))).nonzero()[0]
        if len(indices) > 1:
            mm = torch.abs(grad2[indices[0], indices[1]])
        else:
            mm = torch.abs(grad2[indices])
        n = grad2.size(dim=0)
        if (mm > self.lambda_1):
            grad1 *= self.lambda_1 / mm
        return float("inf") if self.intercept and (torch.abs(grad2[n - 1]) > 1e-6) else 0, grad1, grad2

    def print(self) -> None:
        logger.info(self.getName())

    def lazy_prox(self, input: torch.Tensor, indices: torch.Tensor, eta: float) -> None:
        p = input.size(dim=0)
        # TODO output probablement faux
        # TODO plante surement en 1D
        output = torch.zeros(input.size())
        # Calculate the soft thresholding operation for all elements of the input tensor
        soft_thresholded = input + 0.5 * (torch.abs(input - self.lambda_1) - torch.abs(input + self.lambda_1))

        # Create a mask to select elements along the first dimension based on indices
        mask = torch.zeros(output)
        mask[indices, torch.arange(output.size(1))] = 1

        # Apply the mask to select elements along the first dimension and update them with soft thresholded values
        output.masked_scatter_(mask.bool(), soft_thresholded)
        if (self.intercept):
            if len(output.shape) == 1:
                n = input.size(dim=0)
                output[n - 1] = input[n - 1]
            else:
                p = input.size(dim=1)
                output[:, p - 1] = input[:, p-1]

        return output

    def is_lazy(self) -> bool:
        return True

    def getName(self) -> str:
        return "L1 regularization"



import os
import torch
import numpy as np

EPSILON = 10e-10

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
NUMBER_OPTIM_PROCESS_INFO = 6

datatype = os.environ.get('CYANURE_DATATYPE')

if datatype is not None and datatype == "float64":
    TENSOR_TYPE = torch.float64
    ARRAY_TYPE = np.float64
else:
    TENSOR_TYPE = torch.float32
    ARRAY_TYPE = np.float32



import torch

from cyanure_gpu.constants import DEVICE


class ProblemParameters:
    lambda_1: float
    lambda_2: float
    lambda_3: float
    intercept: bool
    regul: str
    loss: str

    def __init__(self, lambda_1: float = 0, lambda_2: float = 0, lambda_3: float = 0, intercept: bool = False,
                 regul: str = "NONE", loss: str = "SQUARE"):
        self.lambda_1 = torch.tensor([lambda_1], device=DEVICE)
        self.lambda_2 = torch.tensor([lambda_2], device=DEVICE)
        self.lambda_3 = torch.tensor([lambda_3], device=DEVICE)
        self.intercept = intercept
        self.regul = regul
        self.loss = loss
